﻿// textwnd.as
/*
  Тут часть кода, работающая с текстовыми окнами
*/
TextWnd@ activeTextWnd;

// Это интерфейс обработчика событий в тексте.
// В-зависимости от назначенного текстовому документу расширения (встроенный язык, язык запросов и т.п.)
// создается соответствующий расширению текстовый процессор. Именно он и обрабатывает события в текстовом окне
interface TextProcessor
{
    // Подключение нового окна редактора
    void connectEditor(TextWnd@ editor);
    // Обработка после WM_CHAR
    void afterChar(TextWnd@ editor, uint16 symbol);
    // Обработка при WM_KEYDOWN
    void onKeyDown(TextWnd@ editor, uint wParam, uint lParam);
};

// Просто базовый интерфейс для хранения текстовым процессором данных для конкретного редактора
interface EditorData {};

// Реализация пустого текстового процессора, ничего не делает
class EmptyTextProcessor : TextProcessor {
    void connectEditor(TextWnd@ editor) { @editor.editorData = null; }
    void afterChar(TextWnd@ editor, uint16 symbol) {}
    void onKeyDown(TextWnd@ editor, uint wParam, uint lParam) {}
};

/*
    Основная функция вставки текста в текстовый редактор.
    Вставляет в текущее положение редактора (как-будто нажали Ctrl+V)
    При необходимости дополняет многострочный текст нужным отступом,
    взависимости от вставляемого положения.
    При необходимости сбрасывает кэш парсера строк 1С.
    В-принципе, необходимость возникает всегда :)
    иначе очень весёлые глюки. Хотя когда нужно сделать несколько вставок подряд,
    кэш можно сбросить только после последней вставки.
*/
void insertInSelection(ITextEditor@ pEditor, TextManager@ pTM, ITextManager@ pITM, const string& txt, bool bClearCache, bool needIndent)
{
    string text(txt);
    TextPosition tpStart, tpCaretAfter;
    pEditor.getSelection(tpStart, tpCaretAfter, false);
    // Далее надо определить конечное положение селекшена, конечное положение каретки,
    // и сделать по необходимости отступы
    if (text.find('\n') < 0) {  // Однострочный текст, с ним проще
        tpCaretAfter = tpStart;
        int caret = text.find(symbCaret);
        if (caret >= 0) {
            tpCaretAfter.col += caret;
            text.remove(caret);
        } else
            tpCaretAfter.col += text.length;
    } else {
        if (needIndent) {
            // Формируем отступ. Для этого к каждой строке вставляемого текста, начиная со второй
            // нужно добавить такой-же отступ, как у строки в редакторе
            v8string wline;
            pTM.getLine(tpStart.line, wline);   // получили существующую строку текста в формате 1С
            string indent = wline.str.rtrim("\r\n").padRight(' ', tpStart.col - 1).match(indentRex).text(0, 0);
            if (indent.length > 0)
                text.replace("\n", "\n" + indent);
        }
        // Находим конечное положение
        tpCaretAfter = tpStart;
        for (int i = 0, l = text.length; i < l; i++) {
            uint16 symbol = text[i];
            if ('\n' == symbol) {
                tpCaretAfter.line++;
                tpCaretAfter.col = 1;
            } else if (symbCaret == symbol) {
                text.remove(i);
                break;
            } else
                tpCaretAfter.col++;
        }
    }
    // Вставляем текст
    pEditor.setSelectionText(text);
    pEditor.setCaretPosition(tpCaretAfter, false);
    pEditor.scrollToCaretPos();

    if (bClearCache) {
        ITextParserCache@ pCache = cast<IUnknown>(pITM);
        //if (pCache !is null)
            pCache.clearCache();
    }
    // Обновляем все другие окна этого текста
    ITextManager_Operations@ vu = cast<IUnknown>(pITM);
    //if (vu !is null)
        vu.updateAllViews();
}

TextProcessor@ createTextProcessor(const Guid& g)
{
    if (g == gTextExtModule || g == gTextExtModule1) {
        return ModuleTextProcessor();
    }
    return EmptyTextProcessor();
}

TextDocStorage textDocStorage;

class TextDocStorage
{
    array<TextDoc@> openedDocs;

    TextDoc@ find(ITextManager@ itm)
    {
        for (uint idx = 0, size = openedDocs.length(); idx < size; idx++) {
            if (openedDocs[idx].itm is itm)
                return openedDocs[idx];
        }
        return null;
    }
    TextDoc@ find(TextManager@ tm)
    {
        for (uint idx = 0, size = openedDocs.length(); idx < size; idx++) {
            if (openedDocs[idx].tm is tm)
                return openedDocs[idx];
        }
        return null;
    }
    void store(TextDoc@ tdoc) { openedDocs.insertLast(@tdoc); }
    void remove(TextDoc@ doc)
    {
        for (uint idx = 0, size = openedDocs.length(); idx < size; idx++) {
            if (openedDocs[idx] is doc) {
                openedDocs.removeAt(idx);
                break;
            }
        }
    }
};

// Вызывается из перехвата при создании окна текстового редактора
void onCreateTextWnd(IWindowView@ pWnd, bool bControl)
{
    if (bControl) {
        IControlDesign@ ctrl = cast<IUnknown>(pWnd);
        if (ctrl !is null && ctrl.getMode() != ctrlRunning)
            return; // С контролами на формах не работаем
    }
    TextWnd wnd(pWnd, bControl);
    ITextManager@ itm;
    wnd.ted.getITextManager(itm);
    TextDoc@ tdoc = textDocStorage.find(itm);
    if (tdoc is null) {
        //Print("Text doc not found");
        @tdoc = TextDoc();
        @tdoc.itm = itm;
        @tdoc.tm = itm.getTextManager();
        ITextManager_Operations@ to = cast<IUnknown>(itm);
        Guid textExtender;
        to.getExtenderCLSID(textExtender);
        @tdoc.tp = createTextProcessor(textExtender);
        textDocStorage.store(tdoc);
    } //else Print("Found old text doc");
    tdoc.attachView(wnd);
}

// Вызывается из перехвата при смене расширения текстового редактора
// (т.е. в меню текст выбрали тип текста - Встроеный язык, язык запросов, и т.п.)
void changeTextExtender(ITextManager@ itm, const Guid& textExtender)
{
    TextDoc@ tdoc = textDocStorage.find(itm);
    if (tdoc !is null)
        tdoc.setTextExtenderType(textExtender);
}

// Класс, который хранит различную инфу о текстовом документе.
// Важно, у одного текстового документа может быть открыто несколько окон
class TextDoc
{
    ITextManager@   itm;
    TextManager@    tm;
    TextProcessor@  tp;
    array<TextWnd@> views;

    void attachView(TextWnd@ wnd)
    {
        @wnd.textDoc = this;
        views.insertLast(@wnd);
        tp.connectEditor(wnd);
    }
    void detachView(TextWnd@ wnd)
    {
        if (wnd is activeTextWnd)
            @activeTextWnd = null;
        for (uint idx = 0, size = views.length(); idx < size; idx++) {
            if (views[idx] is wnd) {
                views.removeAt(idx);
                break;
            }
        }
        if (views.length() == 0) { // Закрыли последнее окно
            textDocStorage.remove(this);
        }
    }
    void setTextExtenderType(const Guid& g)
    {
        @tp = createTextProcessor(g);
        for (uint k = 0, m = views.length(); k < m; k++)
            tp.connectEditor(views[k]);
    }
};

class TextWnd
{
    ASWnd@          wnd;            // сабклассер текстового окна, перенаправляет указанные события оконной процедуры в скриптовый обработчик
    HWND            hWnd;           // WinAPI хэндл окна
    ITextEditor@    ted;            // Интерфейс редактора 1С
    TextDoc@        textDoc;        // Родительский объект, общий для всех окон одного текста
    EditorData@     editorData;     // Данные, нужные для текстового процессора
    bool            isControl;      // Является ли окно тектовым контролом или полноценным view
    
    TextWnd(IWindowView@ v, bool isCtrl)
    {
        isControl = isCtrl;
        hWnd = v.hwnd();
        @wnd = attachWndToFunction(hWnd, WndFunc(this.WndProc),
                    array<uint> = { WM_KEYDOWN, WM_CHAR, WM_DESTROY, WM_SETFOCUS, WM_KILLFOCUS });
        @ted = cast<IUnknown>(v);
    }
    // Это функция обработчик оконных сообщений сабклассированного окна
    uint WndProc(uint msg, uint w, uint l)
    {
        switch (msg) {
        case WM_SETFOCUS:
            @activeTextWnd = this;
            break;
        case WM_KILLFOCUS:
            @activeTextWnd = null;
            break;
        case WM_DESTROY:
            if (textDoc !is null)
                textDoc.detachView(this);
            break;
        case WM_KEYDOWN:
            return onKeyDown(w, l);
        case WM_CHAR:
            return onChar(w);
        }
        return wnd.doDefault();
    }
    ~TextWnd()
    {
        //Message("destroy TextWnd");
    }
    uint onKeyDown(uint w, uint l)
    {
        textDoc.tp.onKeyDown(this, w, l);
        return wnd.doDefault();
    }
    uint onChar(uint16 symb)
    {
        // сначала дадим отработать штатным механизмам
        uint res = wnd.doDefault();
        // Дальше работает текстовый процессор, назначенный этому тексту
        textDoc.tp.afterChar(this, symb);
        return res;
    }
};
